program hmmstrss
  use hmmstr
!------------------------------------------------------------------------
!  HMMSTRSS
!  Read a FASTA multiple protein sequence alignment (or single sequence)
!  Use HMMSTR to calculate the secondary structure prediction.
! 
!  Chris Bystroff, Pat Buck,  08-DEC-2009, RPI, Troy NY
!------------------------------------------------------------------------
!*=========================================================================
!** Copyright 1996-2009 Chris Bystroff
!**
!** Permission is hereby granted to use, execute, copy, distribute, modify,
!** and distribute in modified form this software for non-profit purposes,
!** provided this notice is retained in any and all copies.
!**
!** For for-profit usage, please contact bystrc@rpi.edu
!*=========================================================================
        !** ------------ declarations ---------------
        implicit none
        character(len=8000)             :: aline,bline,seq
        character(len=8000)             :: pline
        character(len=80)               :: arg1
        integer                         :: I,J,K,L,M,N,nseq,ios=0,Q,NN,II
        character(len=200)              :: infile,outfile
        character(len=1),dimension(:),allocatable :: msa
        integer                         :: jarg,ounit,nalign,ncol,nrec,pcol
        real                            :: id,x, y,confid
        real(4)                         :: ss(6)
        real,dimension(:,:),allocatable :: profile
        integer,dimension(:),allocatable :: pdex
        real,dimension(:,:),pointer     :: xgamma
        character(len=100)              :: modelfile
        character(6)                    :: dssp="HEGST_"
        character(1),dimension(6)       :: hel=(/"H","E","H","L","L","L"/)
        logical                         :: pruning=.true.
        !** ------------ functions ---------------
        integer                 :: iargc
        !** ------------- initialization ---------------
        jarg = iargc()
        if (jarg<3) then
          write(*,'("xhmmstrss fasta_file outputfile modelfile [0/1]")')
          write(*,*) "fasta_file = MSA in FASTA format, gap char is -"
          write(*,*) "outputfile = Secondary structure with the following fields: "
          write(*,*) "SS: residue_number DSSP HEL confid P(H) P(E) P(G) P(S) P(T) P(_)"
          write(*,*) "where residue_number is the column in the MSA"
          write(*,*) "DSSP (Dictionary of Secondary Structure for Proteins) ="//dssp
          write(*,*) "HEL (3-state reduction of DSSP) = H (helix) E (extended) or L."
          write(*,*) "modelfile = HMMSTR model R file. (model_R.hmm)"
          write(*,*) "0/1 = prune gap chars (1) or not (0). Default is 1."
          stop 'hmmstrss.f90 v.  Wed Dec  2 10:00:43 EST 2009'
        endif
        call getarg(1,infile)
        call getarg(2,outfile)
        call getarg(3,modelfile)
        !----- optional args
        if (jarg>=4) then
          call getarg(4,aline)
          read(aline,*,iostat=ios) II
          if (ios/=0) stop '3rd argument must be 0/1'
          pruning = (II==1)
        endif
        !========== open FASTA MSA file ========
        write(*,'(a)') 'Opening FASTA file: '//trim(infile)
        open(11,file=infile,status='old',form='formatted',iostat=ios)
        if (ios/=0) then
          write(0,'(a)') 'File not found: '//trim(infile)
          stop 'FASTA input file not found.'
        endif
        !========== read MSA, FASTA format =========
        !** find out how many total sequences there are. 1 per line.
        !** And get alignment length (number of columns)
        !===========================================
        nalign = 0
        ncol = 0
        n = 0
        do
          read(11,'(a)',iostat=ios) aline
          if (ios/=0) exit
          if (aline(1:1)==">")  then
            if (ncol==0) then
              ncol = n
            elseif (n/=ncol) then
              write(0,'(a)') 'hmmstrss:: ERROR! All sequences in FASTA file must be the same length.'
              write(0,'(a,i4,a,i5)') 'Sequence ',ncol,' Length=',ncol
              write(0,'(a,i4,a,i5)') 'Sequence ',nalign,' Length=',n
              stop 'hmmstrss:: ERROR! All sequences in FASTA file must be the same length.'
            endif
            nalign = nalign + 1
            n = 0
            cycle
          endif
          m = len_trim(aline)
          if (m==0) cycle
          n = n + m
          !!------ diagnostic
          ! write(*,'("sequence ",i4,1x,a)') nalign, trim(aline)
          ! write(*,'("length of line ",i4,i5)') nalign, m
        enddo
        if (ncol==0) then
          ncol = n
        elseif (n/=ncol) then
          write(0,'(a)') 'hmmstrss:: ERROR! All sequences in FASTA file must be the same length.'
          write(0,'(a,i4,a,i5)') 'Sequence ',ncol,' Length=',ncol
          write(0,'(a,i4,a,i5)') 'Sequence ',nalign,' Length=',n
          stop 'hmmstrss:: ERROR! All sequences in FASTA file must be the same length.'
        endif
        !========== allocate memory for MSA =========
        rewind(11)
        allocate(msa(ncol*nalign),stat=ios); if (ios/=0) stop 'hmmstrss:: error allocating memory.'
        allocate(profile(20,ncol)); if (ios/=0) stop 'hmmstrss:: error allocating profile.'
        allocate(pdex(ncol)); if (ios/=0) stop 'hmmstrss:: error allocating pdex.'
        !========== read MSA, FASTA format =========
        n = 0
        nseq = 0
        do
          read(11,'(a)',iostat=ios) aline
          if (ios/=0) exit
          if (aline(1:1)==">")  then
            nseq = nseq + 1
            n = 0
            cycle
          endif
          m = len_trim(aline)
          if (m==0) cycle
          do J=1,m
            n = n + 1
            msa((n-1)*nalign+nseq) = aline(J:J)
          enddo
        enddo
        close(11)
        !---------- save first sequence ---------
        do J=1,ncol
          seq(J:J) = msa((J-1)*nalign+1)
        enddo
        !! diagnostic
        !do I=1,nalign
        !  do J=1,ncol
        !    write(*,'(a1,$)') msa((J-1)*nalign+I) 
        !    if (mod(J,50)==0) write(*,*)
        !  enddo
        !  write(*,'(/,a)') '-----------------------------------------'
        !enddo
        !========== calculate weights, profile, =========
        call writeprofiles(msa,nalign,ncol,profile)
        !========== optionally omit gaps characters  =========
        if (pruning) then
          call prunegap(seq,profile,ncol,pdex,pcol)
          !! diagnostic
          do I=1,pcol
            write(*,'(i4,20f6.3)') I,profile(1:20,I)
          enddo
        else
          pcol = ncol
          do I=1,ncol; pdex(I)=I; enddo
        endif
        !========== calculate gamma  =========
        call hmmstr_gamma(xgamma,pcol,modelfile,profile)
        open(ounit,file=trim(outfile),form='formatted',status='replace',iostat=ios)
        if (ios/=0) stop 'hmmstrss:: error opening output file. Permissions??'
        !========== calculate secondary structure prediction  =========
        aline = " "
        bline = " "
        pline = " "
        do I=1,ncol
          ss(:) = 0.00
          if (pdex(I)==0) then
            ! write(ounit,'("SS:  ",i4,3(1x,a1),f8.4,6f8.3)') I,seq(I:I),&
            !      "_","_",0.000,ss
            cycle
          endif
          II = pdex(I)
          do J=2,modelR%n
            !! modelR%d(hmmstate,dsspstate)
            !! xgamma(hmmstate,residue)
            ss(:) = ss(:) + modelR%d(J,:)*xgamma(J,II)
          enddo
          ss(:) = ss(:)/sum(ss)
          K = maxloc(ss,dim=1)
          L = maxloc(ss,dim=1,mask=(hel/=hel(K)))
          confid = sum(ss,mask=(hel==hel(K)))-sum(ss,mask=(hel==hel(L)))
          confid = max(0.,confid)
          !!---- diagnostic
          !write(*,'("SS:  ",i4,3(1x,a1),f8.4,6f8.3)') I,seq(I:I),&
          !       dssp(K:K),hel(K),confid,ss
          write(ounit,'("SS:  ",i4,3(1x,a1),f8.4,6f8.3)') I,seq(I:I),&
                 dssp(K:K),hel(K),confid,ss
          aline(II:II) = seq(I:I)
          bline(II:II) = dssp(K:K)
          write(pline(II:II),'(i1)') int(10*confid)
        enddo
        close(ounit)
        deallocate(msa)
        deallocate(xgamma)
        deallocate(profile,pdex)
        call release_model()
        !========== output text format =========
        do L=1,1+len_trim(aline)/50
          I = (L-1)*50+1
          J = min(len_trim(aline),L*50)
          write(*,'(a)') aline(I:J)
          write(*,'(a)') bline(I:J)
          write(*,'(a)') pline(I:J)
          write(*,*)
        enddo
        !========== all done. =========
        CONTAINS
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!++++++++ start of contained routines +++++++++++++++++++++++
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!************************************************************************
!        character(len=*) function caps(card)
!        !! Flip a string to uppercase
!        character(len=*),intent(inout) :: card
!        integer                        :: I,icap
!        icap = ichar('A') - ichar('a')
!        do I=1,len(card)
!          if ('a'.le.card(I:I).and.card(I:I).le.'z') &
!          card(I:I) = char(ichar(card(I:I))+icap)
!        enddo
!        caps = card
!        return
!        end function caps
!************************************************************************
!        character(len=*) function lcase(card)
!        !! Flip a string to lowercase
!        character(len=*),intent(inout) :: card
!        integer                        :: I,icap
!        icap = ichar('a') - ichar('A')
!        do I=1,len(card)
!          if ('A'.le.card(I:I).and.card(I:I).le.'Z') &
!          card(I:I) = char(ichar(card(I:I))+icap)
!        enddo
!        lcase = card
!        return
!        end function lcase
!************************************************************************
        subroutine writeprofiles(msa,nalign,nres,profile)
        !-------------------------------------------------------------------
        !** Calculates weights according to Vingron & Argos mismatch method,
        !** with Sander & Schneider (PROTEINS v9) self-consistent condition.
        !-------------------------------------------------------------------
        implicit none
        integer,intent(in)           :: nalign,nres
        character(len=1),intent(in)  :: msa(nalign*nres)
        integer(2)                   :: ident(nalign,nalign)
        real(4),dimension(nalign)    :: sumwts,sumrms,wt,nwt
        integer(4),dimension(nalign) :: iddex,iidex,igdex,nwts
        integer(4)        :: idsum,I,I1,J,K,K1,K2,L,incommon,ires
        real(4)           :: x,y,z
        real(4),intent(out)    :: profile(20,nres)
        real(4):: wsum,csum,wdiff,wtol,gsum,isum
        integer(4)        :: ndex,wcyc,gdex,idex
        character(1)      :: res1(20),aachar,contex,domainID
        !**-----------------------------------------------------
        res1=(/'A','C','D','E','F','G','H','I','K','L', &
                   'M','N','P','Q','R','S','T','V','W','Y'/)
        wtol = 0.02  ! tolerance for self-consistent weights
        do I=1,nalign
          !----------------------------------------------------
          ! Find the number of mismatches (p-distance) between 
          ! any two sequences. ident(I,J) is the p-distance.
          !----------------------------------------------------
          nwts(I) = 0
          sumwts(I) = 0.0
          sumrms(I) = 0.0
          do J=1,I-1
            idsum = 0
            incommon = 0
            do K=1,nres
              K1 = (K-1)*nalign + I
              K2 = (K-1)*nalign + J
              if ((msa(K1) >= 'A').and.(msa(K1) <= 'Z')) then
                if ((msa(K2) >= 'A').and.(msa(K2) <= 'Z')) then
                  incommon = incommon + 1
                  !------  Count mismatches, idsum = mismatches
                  if  (msa(K1) /= msa(K2)) idsum = idsum + 1
                endif
              endif
            enddo
            if (incommon == 0) then  
              !------------------------------------------------------------
              ! incommon can't be zero. NOTE: this happens when two
              ! sequences within the MSA have no overlap. This would produce an
              ! error when generating a phylogenetic tree, but for the
              ! purposes of generating a profile, it is OK.
              !------------------------------------------------------------
              incommon = 1
              idsum = 1
            endif
            x = 100.*real(idsum)/real(incommon)   !! p-distance*100
            ! x = max(0.,real(75-x)/(75.))        !! old correction
            y = 1 - x/75.
            if (y < 0.) y = 1 - 74./75.
            x = -75.*log(y)                       !! Jukes-Cantor distance*100
            ident(I,J) = nint(x)
            ident(J,I) = nint(x)
          enddo
          ident(I,I) = 0    ! by definition, right?
        enddo
        profile = 0.0
        ILOOP: do I=1,nres
          !-----------------------------------------------------
          ! Calculate position-specific weights. Sequences
          ! are weighted differently if the positions 
          ! have different gap configurations. Positions are
          ! weighted the same if they have the same
          ! arrangement of gaps (same set of non-gap sequences).
          ! From Hennikoff & Hennikoff.
          !-----------------------------------------------------
          ndex = 0
          gdex = 0
          idex = 0
          JLOOP: do J=1,nalign
            !-----------------------------------------------
            !** For this position and sequence, ask if it is a 
            !** match, deletion or insertion. Save the sequence
            !** number in iddex, igdex or iidex respectively.
            !** Skip duplicate sequences, for which ident()==0
            !-----------------------------------------------
            I1 = (I-1)*nalign + J
            if (msa(I1) >= 'A'.and.msa(I1) <= 'Z') then
              !-----------------------------------------------
              ! Anything that is not an uppercase AA character
              ! is ignored.
              ! ONLY THIS DATA IS CURRENTLY USED FOR PROFILE CALC
              !-----------------------------------------------
              L = 100
              do K=1,ndex
                L = min(L,ident(J,iddex(K)))
              enddo
              if (L > 0) then
                ndex = ndex + 1
                iddex(ndex) = J
              endif
            elseif (msa(I1) == '-') then ! deletion
              !-----------------------------------------------
              ! Gap character is '-'. This is a deletion
              ! relative to the parent sequence.
              ! THIS DATA IS CURRENTLY NOT USED.
              !-----------------------------------------------
              L = 100
              do K=1,gdex
                L = min(L,ident(J,igdex(K)))
              enddo
              if (L > 0) then
                gdex = gdex + 1
                igdex(gdex) = J
              endif
            elseif (msa(I1) >= 'a'.and.msa(I1) <= 'z') then 
              !-----------------------------------------------
              ! Lowercase AA character in the MSA means insertion
              ! occurs here, and the insert sequence is absent.
              ! If the MSA includes all residues of all sequences
              ! then there will be not insertions.
              ! THIS DATA IS CURRENTLY NOT USED.
              !-----------------------------------------------
              L = 100
              do K=1,idex
                L = min(L,ident(J,iidex(K)))
              enddo
              if (L > 0) then
                idex = idex + 1
                iidex(idex) = J
              endif
            endif
          enddo JLOOP
          !----------------------------------------------------
          !** calculate the self-consistent weight set 
          !** (Sander&Schneider) for this position.
          !----------------------------------------------------
          wt(:) = 1.0        !! initialize to equal weights.
          wdiff = wtol + 1.  !! initialize to > wtol
          wcyc = 0
          do while (wdiff > wtol.and.wcyc < 100)
            wdiff = 0.0  ! sum of weights
            do J=1,ndex
              wsum = 0.0
              do L=1,ndex
                x = wt(L)*real(ident(iddex(J),iddex(L)))
                wsum = wsum + x/100.
              enddo
              nwt(J) = wsum
              wdiff = wdiff + wsum
            enddo
            if (wdiff == 0.) exit
            nwt(1:ndex) = nwt(1:ndex)/wdiff
            wdiff = 0.0      !! sum the avg relative change
            wsum = 0.0
            do J=1,ndex
              wdiff = wdiff + abs(wt(J) - nwt(J))
              wsum = wsum + wt(J) + nwt(J)
              wt(J) = nwt(J)
            enddo
            wcyc = wcyc + 1
            wdiff = wdiff/wsum
          enddo
          !-------------------------------------------------
          ! The following error message signals a bug.
          !-------------------------------------------------
          if (wcyc >= 100) write(0,*) 'hmmstrss:: WARNING! weights not converging!!'
          !-------------------------------------------------
          ! done calculating weight set, sum profile 
          ! for this position
          !-------------------------------------------------
          csum = 0.0
          do J=1,ndex
            wsum = wt(J)
            I1 = (I-1)*nalign + iddex(J)
            ires = 0
            do K=1,20
              !-------------------------------------------------
              ! Find out which amino acid this is.
              !-------------------------------------------------
              if (msa(I1) == res1(K)) ires = K
            enddo
            if (ires /= 0) then
              profile(ires,I) = profile(ires,I) + wsum
              csum = csum + wsum
            else
              !-------------------------------------------------
              ! This is a warning message that the AA was 
              ! unrecognized. It will be ignored in the profile.
              !-------------------------------------------------
              write(*,'(" Unknown AA in sequence ",i4,'// &
              ' ": ",a1)') iddex(J),msa(I1)
            endif
          enddo ! J
          !-------------------------------------------------
          !** OPTIONAL/DIAGNOSTIC: 
          !** output number of seqs at this position, ndex
          !-------------------------------------------------
          ! write(*,*) I, ndex
          if (csum > 0.0) then
            !-------------------------------------------------
            ! OPTIONAL: sum the weights and weights^2 to
            ! later calculate the RMS weight.
            !-------------------------------------------------
            !do J=1,ndex
            !  wsum = wt(J)/csum
            !  sumwts(iddex(J)) = sumwts(iddex(J)) + wsum
            !  sumrms(iddex(J)) = sumrms(iddex(J)) + wsum*wsum
            !  nwts(iddex(J)) = nwts(iddex(J)) + 1
            !enddo ! J
            !-------------------------------------------------
            ! Normalize the AA profile to sum to 1.00
            !-------------------------------------------------
            profile(:,I) = profile(:,I)/csum
          endif
          !-------------------------------------------------
          !** OPTIONAL: output Shannon negative entropy at this position
          !-------------------------------------------------
          !csum = 0.
          !do J=1,20
          !  if (profile(J,I)>0.000) then
          !    csum = csum - profile(J,I)*log(profile(J,I))
          !  endif
          !enddo
          !-------------------------------------------------
        enddo ILOOP
        return
        end subroutine writeprofiles
!************************************************************************
        subroutine prunegap(seq,profile,ncol,pdex,pcol)
          implicit none
          character(len=*),intent(in)           :: seq
          integer,intent(in)                    :: ncol
          integer,intent(out)                   :: pcol
          real,dimension(20,ncol),intent(inout) :: profile
          integer,intent(out),dimension(ncol)   :: pdex
          integer                               :: I,J
          !-------------------------------------------------
          ! Condense profile using only non-gap columns
          !-------------------------------------------------
          pcol = 0
          pdex = 0
          do I=1,ncol
            if (seq(I:I) >= 'A'.and.seq(I:I) <= 'Z') then
              pcol=pcol+1
              pdex(I) = pcol
              profile(1:20,pcol) = profile(1:20,I)
            endif
          enddo
        end subroutine prunegap
!************************************************************************
!******* end of contained routines  *************************************
!************************************************************************
end program hmmstrss
